{------------------------------------------------------------------------------}
{
{                               SYMBOLTABLE.PAS
{
{    Gerncia da Tabela de Smbolos
{
{    Sistema:    DosVox
{    Mdulo:     Interpretador ScriptVox
{    Autor:      Oswaldo Vernet
{    Data:       21/08/2015
{    Alteraes:
{
{------------------------------------------------------------------------------}

unit SYMBOLTABLE;

{------------------------------------------------------------------------------}
{                             I N T E R F A C E
{------------------------------------------------------------------------------}

interface

uses
    lex, sysUtils;

{**************************** A Tabela de Smbolos ****************************}

type
    PSymtbEntry         = ^SymtbEntry;

    SymbolType          = (
                                S_UNDEF,                                        { Indefinido }
                                S_COMM,                                         { Comando }
                                S_IF, S_UIF, S_ELSE, S_ELIF,                    { Comandos SE, SENO e SENO SE }
                                S_CFUNC,                                        { Comando FUNO }
                                S_LOOP,                                         { Comandos iterativos }
                                S_BREAK, S_CONT,                                { Comandos QUEBRA e CONTINUA }
                                S_END,                                          { Comando FIM }
                                S_ROT,                                          { Rtulo }
                                S_VAR,                                          { Varivel }
                                S_MOD,                                          { Descritor de mdulo }
                                S_NFUNC,                                        { Funo nativa }
                                S_UFUNC                                         { Funo definida pelo usurio }
                          );

    SymbolTypeSet       = set of SymbolType;
    FunExcep            = ( FE_SUS, FE_VARARG );
    ExecuteCmd          = function : boolean;
    ExecuteFun          = function (exp : pointer; arg : pointer) : boolean;

    SymtbEntry          = record
                              id        : string;                               { Identificador }

                              script    : integer;                              { Escopo }
                              scope     : PSymtbEntry;

                              snext     : PSymtbEntry;                          { Prximo  na lista de smbolos }
                              hnext     : PSymtbEntry;                          { Prximo  na cadeia de colises }
                              hprev     : PSymtbEntry;                          { Anterior na cadeia de colises }

                              case typeof : SymbolType of                       { Tipo do smbolo }
                                  S_COMM,
                                  S_IF, S_UIF, S_ELSE, S_ELIF,
                                  S_CFUNC,
                                  S_LOOP,
                                  S_BREAK, S_CONT,
                                  S_END:        ( exec     : ExecuteCmd );      { Processamento do comando }
                                  S_ROT:        ( line     : integer );         { Linha em que foi definido o rtulo }
                                  S_VAR, S_MOD: ( val      : pointer;           { Valor (s para variveis globais) }
                                                  offset   : integer );         { Deslocamento (para variveis na pilha) }
                                  S_NFUNC:      ( fexec    : ExecuteCmd;        { Processamento da funo nativa }
                                                  narg     : integer;           { Nmero de argumentos }
                                                  excep    : set of FunExcep ); { Excees }
                                  S_UFUNC:      ( start,                        { Linha inicial }
                                                  finish,                       { Linha final }
                                                  nargs,                        { Nmero de argumentos }
                                                  local    : integer );         { Nmero de variveis locais }
                          end;

const
    SymbolTypeToStr     : array [SymbolType] of string[20] =                    { Nome de cada um dos tipos de smbolos }
                          (
                                'indefinido', 'comando', 'comando',  'comando',
                                'comando',    'comando', 'comando',  'comando',
                                'comando',    'comando', 'comando',  'rtulo',
                                'varivel',   'referncia a mdulo', 'funo nativa',
                                'funo'
                          );

var
    UnitaryIfPtr        : PSymtbEntry;                                          { Referncia para o comando SE unitrio }
    AssignmentPtr       : PSymtbEntry;                                          { Referncia para o comando := }
    EvaluationPtr       : PSymtbEntry;                                          { Referncia para o comando := }
    ReturnCmdPtr        : PSymtbEntry;                                          { Referncia para o comando RETORNA }

procedure defineStaticSymbol         (q : PSymtbEntry);
function  defineLocalVariable        (s : integer; sc : PSymtbEntry; id : string; offset : integer) : PSymtbEntry;
function  defineLabel                (s : integer; sc : PSymtbEntry; id : string; line : integer) : boolean;
function  getVariable                (s : integer; sc : PSymtbEntry; id : string) : PSymtbEntry;
function  getSymbol                  (s : integer; sc : PSymtbEntry; id : string; define : boolean) : PSymtbEntry;
function  getLabel                   (s : integer; sc : PSymtbEntry; id : string): integer;
function  getCommand                 (id : string) : PSymtbEntry;
procedure removeAllSymbols           (s : integer);
function  sizeOfLargestCollisionList () : integer;
function  getFirstSymbol             () : PSymtbEntry;
function  getNextSymbol              (p : PSymtbEntry) : PSymtbEntry;

{------------------------------------------------------------------------------}
{                           I M P L E M E N T A   O
{------------------------------------------------------------------------------}

implementation

const
    HASH_SIZE     = 4001;             { Nmero primo bem grande }

type
    HASH_RANGE    = 0 .. HASH_SIZE - 1;
    HashTableType = array [HASH_RANGE] of PSymtbEntry;

var
    hashTable     : HashTableType;    { Tabela "hash" }

    firstSymbol   : SymtbEntry;       { Incio e fim da tabela de smbolos }
    lastSymbol    : PSymtbEntry;

{------------------------------------------------------------------------------}
{      calcula o "hash" de um identificador
{------------------------------------------------------------------------------}

function hash (id : string) : HASH_RANGE;
var
    i, len  : integer;
    sum     : int64;
begin
    len := length (id);
    sum := 0;

    for i := 1 to len do
        sum := (sum shl 1) xor ord (id[i]);

    hash := Abs (sum) mod HASH_SIZE
end;

{------------------------------------------------------------------------------}
{      define um smbolo esttico
{------------------------------------------------------------------------------}

procedure defineStaticSymbol (q : PSymtbEntry);
var
    p : ^PSymtbEntry;                 { Ponteiro duplo! }
begin
    { Smbolos estticos so inseridos apenas na tabela hash.
      O n j vem pronto, no  preciso criar! }

    p         := @hashTable[hash (q^.id)];

    q^.script := -1;                  { So independentes de script }
    q^.scope  := NIL;
    q^.snext  := NIL;
    q^.hnext  := p^;                  { Insere no incio da lista de coliso }
    q^.hprev  := NIL;

    if p^ <> NIL then
        p^^.hprev := q;

    p^ := q
end;

{------------------------------------------------------------------------------}
{      busca uma varivel (ou funo)
{------------------------------------------------------------------------------}

function getVariable (s : integer; sc : PSymtbEntry; id : string) : PSymtbEntry;
var
    local, global, p : PSymtbEntry;
begin
    local := NIL; global := NIL;

    p := hashTable[hash (id)];        { Incio da lista de coliso }

    while p <> NIL do                 { Percorre toda a lista de colises }
    begin
        if (p^.id = id) and ((p^.script < 0) or (p^.script = s)) then
        begin
            if p^.scope = NIL then
                global := p
            else if p^.scope = sc then
                local := p
        end;

        p := p^.hnext
    end;

    if local = NIL then local := global;

    getVariable := local
end;

{------------------------------------------------------------------------------}
{      busca um smbolo, definindo-o se requerido
{------------------------------------------------------------------------------}

function getSymbol (s : integer; sc : PSymtbEntry; id : string; define : boolean) : PSymtbEntry;
var
    h           : integer;
    p, q, ant   : PSymtbEntry;
begin
    h := hash (id);

    ant := NIL;
    p   := hashTable[h];              { Busca na lista de colises }

    while p <> NIL do
    begin
        if (p^.id = id) and ((p^.typeof in [S_COMM..S_END,S_NFUNC]) or (((p^.script < 0) or (p^.script = s)) and (p^.scope = sc))) then
        begin
            getSymbol := p;           { Achou! }
            exit
        end;

        ant := p;                     { Colidiu: tenta o prximo }
        p   := p^.hnext
    end;

    q := NIL;                         { Prepara o retorno }

    if define then                    { O smbolo  novo e deve ser definido }
    begin
        try NEW (q) except end;

        if q <> NIL then
        begin
            if (sc = NIL) and (id[1] = '$') then
                q^.script := -1       { Superglobal }
            else
                q^.script := s;

            q^.id     := id;
            q^.scope  := sc;
            q^.snext  := NIL;
            q^.typeof := S_UNDEF;     { Inicialmente no sabemos o tipo }

            q^.val    := NIL;         { ... tampouco o valor }

            q^.hnext  := p;           { Insere na lista de colises }
            q^.hprev  := ant;

            if p <> NIL then
                p^.hprev      := q;

            if ant = NIL then
                hashTable[h]  := q
            else
                ant^.hnext    := q;

            lastSymbol^.snext := q;   { Insere tambm na lista de smbolos }
            lastSymbol        := q
        end
        else begin
            errorMessage ('Memria insuficiente')
        end
    end;

    getSymbol := q
end;

{------------------------------------------------------------------------------}
{      define uma varivel local (ou parmetro)
{------------------------------------------------------------------------------}

function defineLocalVariable (s : integer; sc : PSymtbEntry; id : string; offset : integer) : PSymtbEntry;
var
    p : PSymtbEntry;
begin
    p := getSymbol (s, sc, id, true);

    if not (p^.typeof in [S_VAR,S_UNDEF]) then
    begin
        errorMessage ('"' + id + '"  identificador de ' + SymbolTypeToStr[p^.typeof] + ', no pode ser uma varivel local ou parmetro');
        defineLocalVariable := NIL;
        exit
    end;

    p^.typeof := S_VAR;
    p^.offset := offset;

{   scWriteln ('Varivel local: ' + id + ', desloc = ' + intToStr (desloc)); }

    defineLocalVariable := p;
end;

{------------------------------------------------------------------------------}
{      define um rtulo
{------------------------------------------------------------------------------}

function defineLabel (s : integer; sc : PSymtbEntry; id : string; line : integer) : boolean;
var
    p : PSymtbEntry;
begin
    defineLabel := false;

    p := getSymbol (s, sc, id, true);

    if p = NIL then exit;

    if p^.typeof = S_UNDEF then
    begin
        p^.typeof     := S_ROT;       { Define o rtulo }
        p^.line       := line;
        defineLabel   := true
    end
    else if p^.typeof = S_ROT then
    begin
        errorMessage ('Rtulo "' + id + '" j foi definido na linha ' + intToStr (p^.line))
    end
end;

{------------------------------------------------------------------------------}
{      busca um rtulo
{------------------------------------------------------------------------------}

function getLabel (s : integer; sc : PSymtbEntry; id : string): integer;
var
    p : PSymtbEntry;
begin
    if (length (id) = 0) or (id[1] <> '@') then
        id := '@' + id;

    p := getSymbol (s, sc, lex.normalize (id), false);

    if (p = NIL) or (p^.typeof <> S_ROT) then
        getLabel := -1
    else
        getLabel := p^.line
end;

{------------------------------------------------------------------------------}
{      busca um comando
{------------------------------------------------------------------------------}

function getCommand (id : string) : PSymtbEntry;
var
    p : PSymtbEntry;
begin
    p := hashTable[hash (id)];        { Incio da lista de coliso }

    while p <> NIL do                 { Busca na lista de colises }
    begin
        if (p^.id = id) and (p^.typeof in [S_COMM..S_END]) then
            break;

        p := p^.hnext
    end;

    getCommand := p
end;

{------------------------------------------------------------------------------}
{      apaga os smbolos de um script
{------------------------------------------------------------------------------}

procedure removeAllSymbols (s : integer);
var
    ant, p  : PSymtbEntry;
    h       : integer;
begin
    ant := @firstSymbol; p := ant^.snext;

    while p <> NIL do
    begin
        if p^.script = s then
        begin
            if p^.typeof = S_VAR then
            begin
                if (p^.scope = NIL) and (p^.offset = 0) then
                    // liberaExpr (p^.val.exp)
            end;

            h := hash (p^.id);

            if p^.hprev = NIL then
                hashTable[h] := p^.snext
            else
                p^.hprev^.hnext := p^.hnext;

            if p^.hnext <> NIL then
                p^.hnext^.hprev := p^.hprev;

            ant^.snext := p^.snext;
            p := ant^.snext
        end
        else begin
            ant := p;
            p := p^.snext
        end
    end
end;

{------------------------------------------------------------------------------}
{      fornece acesso ao primeiro smbolo
{------------------------------------------------------------------------------}

function getFirstSymbol : PSymtbEntry;
begin
    getFirstSymbol := firstSymbol.snext
end;

{------------------------------------------------------------------------------}
{      fornece acesso ao smbolo seguinte
{------------------------------------------------------------------------------}

function getNextSymbol (p : PSymtbEntry) : PSymtbEntry;
begin
    if p = NIL then
        getNextSymbol := NIL
    else
        getNextSymbol := p^.snext
end;

{------------------------------------------------------------------------------}
{      retorna o tamanho da maior lista de colises
{------------------------------------------------------------------------------}

function sizeOfLargestCollisionList : integer;
var
    n, h, m : integer;
    p       : PSymtbEntry;
begin
    m := -1;

    for h := 0 to HASH_SIZE - 1 do
    begin
        n := 0;
        p := hashTable[h];

        while p <> NIL do
        begin
            INC (n);
            p := p^.hnext
        end;

        if n > m then m := n
    end;

    sizeOfLargestCollisionList := m
end;

{------------------------------------------------------------------------------}
{               inicializa a tabela hash
{------------------------------------------------------------------------------}

procedure initHashTable;
var
    i : integer;
begin
    for i := 0 to HASH_SIZE - 1 do
        hashTable[i] := NIL
end;

{------------------------------------------------------------------------------}
{             inicializa a tabela de smbolos
{------------------------------------------------------------------------------}

procedure inicializaTabelaDeSimbolos;
begin
    firstSymbol.snext := NIL;         { O primeiro n no contm informao }
    lastSymbol        := @firstSymbol
end;

begin
    initHashTable;
    inicializaTabelaDeSimbolos
end.




